#include <filesystem>
#include <iostream>

#include <fggl/gfx/window.hpp>
#include <fggl/gfx/camera.hpp>
#include <fggl/input/camera_input.h>

#include <fggl/gfx/compat.hpp>
#include <fggl/gfx/ogl/compat.hpp>

#include <fggl/data/procedural.hpp>
#include <fggl/data/storage.hpp>
#include <fggl/util/chrono.hpp>
#include <fggl/ecs3/ecs.hpp>

#include <fggl/debug/debug.h>
#include <imgui.h>

constexpr bool showNormals = false;

// prototype of resource discovery
void discover(const std::filesystem::path& base) {

	std::vector< std::filesystem::path > contentPacks;
	for ( auto& item : std::filesystem::directory_iterator(base) ) {

		// content pack detection
		if ( std::filesystem::is_directory( item ) ) {
			auto manifest = item.path() / "manifest.yml";
			if ( std::filesystem::exists( manifest ) ) {
				contentPacks.push_back( item.path() );
			}
		}
		
	}

	// what did we find?
	std::cerr << "found pack(s): " << std::endl;
	for ( auto& pack : contentPacks ) {
		std::cerr << pack << std::endl;
	}

}

enum camera_type { cam_free, cam_arcball };
camera_type cam_mode = cam_free;

//TODO proper input system
using namespace fggl::input;
using InputManager = std::shared_ptr<fggl::input::Input>;

void process_camera(fggl::gfx::Window& window, fggl::ecs3::World& ecs, const InputManager& input) {
    auto cameras = ecs.findMatching<fggl::gfx::Camera>();
    fggl::ecs3::entity_t cam = cameras[0];

	auto camTransform = ecs.get<fggl::math::Transform>(cam);
	auto camComp = ecs.get<fggl::gfx::Camera>(cam);

	const glm::vec3 dir = ( camTransform->origin() - camComp->target );
	const glm::vec3 forward = glm::normalize( dir );

	// scroll wheel
	glm::vec3 motion(0.0f);
	float delta = input->mouse.axis( fggl::input::MouseAxis::SCROLL_Y );
	if ( (glm::length( dir ) < 25.0f && delta < 0.0f) || (glm::length( dir ) > 2.5f && delta > 0.0f) )
		motion -= (forward * delta);
	camTransform->origin( camTransform->origin() + motion );

	if ( cam_mode == cam_arcball || input->mouse.button( fggl::input::MouseButton::MIDDLE ) ) { 
		fggl::input::process_arcball(ecs, *input, cam);
	} else if ( cam_mode == cam_free ) {
		fggl::input::process_freecam(ecs, *input, cam);
	}
	fggl::input::process_edgescroll( ecs, *input, cam );
}

int main(int argc, char* argv[]) {

	// setup ECS
    fggl::ecs3::TypeRegistry types;
    fggl::ecs3::ModuleManager modules(types);

    fggl::ecs3::World ecs(types);

    auto inputs = std::make_shared<fggl::input::Input>();
    auto glfwModule = modules.load<fggl::gfx::ecsGlfwModule>(inputs);

	fggl::gfx::Window win{};
	win.title("FGGL Demo");
	win.fullscreen( true );

	// storage API
	fggl::data::Storage storage;
	//discover( storage.resolvePath(fggl::data::Data, "../../packs") );

	// Opengl APIs
    auto glModule = modules.load<fggl::gfx::ecsOpenGLModule>(win, storage);

	fggl::gfx::loadPipeline(glModule, "unlit", false);
	fggl::gfx::loadPipeline(glModule, "phong", false);
	fggl::gfx::loadPipeline(glModule, "normals", false);

	// debug layer
	fggl::debug::DebugUI debug(win);
	debug.visible(true);

	// create ECS
    types.make<fggl::math::Transform>();

    // create camera using strings
    {
        auto prototype = ecs.create(false);
        ecs.add(prototype, types.find(fggl::math::Transform::name));
        ecs.add(prototype, types.find(fggl::gfx::Camera::name));
        ecs.add(prototype, types.find(fggl::input::FreeCamKeys::name));

        auto camTf = ecs.get<fggl::math::Transform>(prototype);
        camTf->origin( glm::vec3(0.0f, 3.0f, 3.0f) );

        auto cameraKeys = ecs.get<fggl::input::FreeCamKeys>(prototype);
        cameraKeys->forward = glfwGetKeyScancode(GLFW_KEY_W);
        cameraKeys->backward = glfwGetKeyScancode(GLFW_KEY_S);
        cameraKeys->left = glfwGetKeyScancode(GLFW_KEY_A);
        cameraKeys->right = glfwGetKeyScancode(GLFW_KEY_D);
        cameraKeys->rotate_cw = glfwGetKeyScancode(GLFW_KEY_Q);
        cameraKeys->rotate_ccw = glfwGetKeyScancode(GLFW_KEY_E);
    }

    // create building prototype
    fggl::ecs3::entity_t bunker;
    {
        bunker = ecs.create(true);
        ecs.add(bunker, types.find(fggl::math::Transform::name));

        // mesh
        int nSections = 2;
        constexpr float HALF_PI = M_PI / 2.0f;
        constexpr char shader[] = "phong";

        fggl::data::Mesh mesh;
        for (int j=-(nSections/2); j<=nSections/2; j++) {
            const auto shapeOffset = glm::vec3( 0.0f, 0.0f, j * 1.0f );

            const auto cubeMat = glm::translate( fggl::math::mat4( 1.0f ) , shapeOffset );
            const auto leftSlope = fggl::math::modelMatrix(
                    glm::vec3(-1.0f, 0.0f, 0.0f) + shapeOffset,
                    glm::vec3( 0.0f, -HALF_PI, 0.0f) );
            const auto rightSlope = fggl::math::modelMatrix(
                    glm::vec3( 1.0f, 0.0f, 0.0f) + shapeOffset,
                    glm::vec3( 0.0f, HALF_PI, 0.0f) );

            fggl::data::make_cube( mesh, cubeMat );
            fggl::data::make_slope( mesh, leftSlope );
            fggl::data::make_slope( mesh, rightSlope );
        }
        mesh.removeDups();

        fggl::gfx::StaticMesh staticMesh{mesh, shader};
        ecs.set<fggl::gfx::StaticMesh>(bunker, &staticMesh);
    }

	auto floorEnt = ecs.create(false);
	{
		ecs.add<fggl::math::Transform>( floorEnt );
		fggl::data::Mesh mesh = fggl::data::make_quad_xz();

        fggl::gfx::StaticMesh sMesh{mesh, "phong"};
		ecs.set<fggl::gfx::StaticMesh>(floorEnt, &sMesh);
	}

    int nCubes = 3;
	for ( int i=0; i<nCubes; i++ ) {
        auto bunkerClone = ecs.copy(bunker);
        auto result = ecs.get<fggl::math::Transform>(bunkerClone);
		result->origin( glm::vec3( i * 5.0f, 0.0f, 0.0f) );
	}

	bool gamepadWindow = true;

	fggl::util::Timer time{};
	time.frequency( glfwGetTimerFrequency() );
	time.setup( glfwGetTimerValue() );

	while( !win.closeRequested() ) {

		//
		// Setup setup
		//
		time.tick( glfwGetTimerValue() );
		inputs->frame( time.delta() );

		glfwModule->context.pollEvents();
		debug.frameStart();

		//
		// update step
		// 
		process_camera(win, ecs, inputs);

        // ECS3 inspector
        ImGui::Begin("Entities");
        auto entityItr = ecs.all();
        for (auto& entity : entityItr) {
            std::string label = "entity-" + std::to_string(entity);
            if ( ImGui::TreeNode(label.c_str()) ){
                auto entComp = ecs.getComponents(entity);
                for ( auto comp : entComp ){
                    auto meta = types.meta(comp);
                    ImGui::Text("%s (%d) - %lu bytes", meta->name(), comp, meta->size());
                }
                ImGui::TreePop();
            }
        }
        ImGui::End();

		// imgui gamepad debug
        if ( gamepadWindow ) {
            auto &gamepads = inputs->gamepads;
            ImGui::Begin("GamePad", &gamepadWindow);
            for (int i = 0; i < 16; i++) {
                std::string title = gamepads.name(i);

                bool present = gamepads.present(i);
                if (ImGui::TreeNode(title.c_str())) {
                    ImGui::Text("present: %s", present ? "yes" : "no");

                    if (present) {

                        if (ImGui::TreeNode("buttons##2")) {
                            for (auto &btn: fggl::input::GamepadButtonsMicrosoft) {
                                ImGui::Text("%s: %i %i %i", btn.name,
                                            gamepads.button(i, btn.id),
                                            gamepads.buttonPressed(i, btn.id),
                                            gamepads.buttonReleased(i, btn.id)
                                );
                            }
                            ImGui::TreePop();
                        }

                        if (ImGui::TreeNode("axes##2")) {
                            for (auto &axis: fggl::input::GamepadAxes) {
                                ImGui::Text("%s: %f %f", axis.name,
                                            gamepads.axis(i, axis.id),
                                            gamepads.axisDelta(i, axis.id)
                                );

                            }
                            ImGui::TreePop();
                        }

                    }

                    ImGui::TreePop();
                    ImGui::Separator();
                }

            }
            ImGui::End();
        }

		//
		// render step
		// 
		win.activate();
		glModule->ogl.clear();

		// model rendering system
		fggl::gfx::renderMeshes(glModule, ecs, time.delta());
		debug.draw();

		// swap the windows - frame rendering over
		win.swap();
	}

	return 0;
}
